#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import boto3
import os
import json
import logging
import webhook_authenticator

SMS_TEMPLATE = 'Hello %s, this is a confirmation of your form submission #%s on %s.'


logger = logging.getLogger()
logger.setLevel(logging.INFO)


def send_sms(phone_number, message):
    """
    :param phone_number: phone number in E.164 format
    :param message: SMS message text
    """
    client = boto3.client('sns')
    client.publish(PhoneNumber=phone_number, Message=message)


def format_number(number, country_code=1):
    """
    Format phone number in E.164 international number format. https://en.wikipedia.org/wiki/E.164
    :param number: phone number
    :param country_code: country dialing code, defaults to 1 (US, Canada,...)
    """
    num = [c for c in number if c.isdigit()]
    num.insert(0, '+%s' % country_code)
    return ''.join(num)


def lambda_handler(event, context):
    """
    The code executed each time your form is submitted
    :param event: the request info passed to Lambda, including the webhook notification payload
    :param context: not used
    """
    # Reconstruct webhook URL so it doesn't have to be configures as an environment variable
    url = '%s://%s%s' % (
        event['headers']['X-Forwarded-Proto'],
        event['multiValueHeaders']['Host'][0],
        event['requestContext']['path']
    )
    auth = webhook_authenticator.Authenticator(os.environ['FORMSMARTS_WEBHOOK_KEY'], url, event['body'])

    # Verify request is from FormSmarts
    if auth.verify_request(event['headers']['Authorization']):
        notif = json.loads(event['body'])
        message = SMS_TEMPLATE % (
            # Load FormSmarts webhook notification JSON payload
            notif['fields'][0]['field_value'].title(),  # The person's name is the value of the fist field
            notif['fs_ref_num'],  # FormSmarts reference number of the form entry
            notif['api_date']  # UTC time and date the form was submitted
        )
        # Phone number is the value of the second field
        number = format_number(notif['fields'][1]['field_value'])
        send_sms(number, message)
        logger.info('Confirmed form entry #%s to %s.' % (notif['fs_ref_num'], number))
        return {'statusCode': 200}  # OK
    else:
        return {'statusCode': 401}  # Authentication failed
