# -*- coding: utf-8 -*-

"""

FormSmarts Webhook Authenticator for Python 3 -- v. 2018-09-11

This tool allows FormSmarts users to check a webhook notifications originates from FormSmarts
before processing the JSON data received in the request body.

https://formsmarts.com/online-form-api-webhook#message-authentication

"""


import urllib.parse
import hmac
import hashlib
import binascii
import json
import time


class Authenticator(object):

    SIGNATURE_TIMEOUT = 15 * 60  # request signature expires after 15 min

    def __init__(self, webhook_key, url, request_body, http_method='POST'):
        """
        :param webhook_key: Your FormSmarts Webhook Key (shown on https://formsmarts.com/account/view)
        :param url: the webhook URL
        :param request_body: the body of the Webhook's HTTP request
        :param http_method: the request's HTTPS method, POST by default
        """
        self._wh_key = webhook_key
        self._url = url
        self._parameters = json.loads(request_body)
        self._http_method = http_method
        self._auth_header = {}
        self._serial_params = {}

    def verify_request(self, authorization_header, check_timestamp=True):
        """
        Return True if the webhook notification is authenticated as originating from FormSmarts, False otherwise.

        :param authorization_header: the Authorization header of the HTTP request
        :param check_timestamp: if True, an AuthenticationError is raise if the signature has expired
        :raise AuthenticationError: if the Authorization header is missing or unparsable or the request has expired.
        """
        auth_header_sign, auth_header = self.authorization_header(authorization_header)
        self.set_authorization_header(auth_header)
        if check_timestamp:
            try:
                timestamp = int(auth_header['oauth_timestamp'])
            except ValueError:
                raise AuthenticationError('Invalid timestamp in Authorization header')
            if int(time.time()) - timestamp > self.SIGNATURE_TIMEOUT:
                raise AuthenticationError(
                    'Request signature has expired, it was generated more than %d seconds ago.' % self.SIGNATURE_TIMEOUT
                )
        return hmac.compare_digest(auth_header_sign, self.signature())

    def set_authorization_header(self, header):
        """
        :param header: the partial OAuth authorization header, without 'oauth_signature'.
        """
        self._auth_header = header

    def base_string(self):
        """
        Build signature base string.
        """
        parts = (
            self._http_method,
            escape(self.normalize_url()),
            escape(self.normalize_parameters()),
        )
        key = '%s&' % self._wh_key
        return key, '&'.join(parts)

    def signature(self):
        """
        Build the signature.
        """
        key, base = self.base_string()
        # calculate hash and return a base-64 encoded string
        hash = hmac.new(bytes(key, 'utf-8'), bytes(base, 'utf-8'), hashlib.sha1)
        return binascii.b2a_base64(hash.digest())[:-1].decode('utf-8')

    def serialize_parameters(self):
        """
        Serialize parameters so they can be sorted in lexicographical order.
        """
        if not self._serial_params:
            for k, v in self._parameters.items():
                self.__serialize(k, v)
        return self._serial_params

    def __serialize(self, key, val):
        if not hasattr(val, '__getitem__') or type(val) == str:
            # not a sequence or if this is a sequence, it's a string
            self._serial_params[key] = val  # normalize None to empty string
            return
        if not hasattr(val, 'keys'):
            # a sequence but not a dictionary: decorate key with index
            for i, v in enumerate(val):
                self.__serialize('_'.join((key, str(i))), v)
            return
        # a dictionary: decorate key with sub-key name
        for k, v in val.items():
            self.__serialize('_'.join((key, k)), v)

    def normalize_parameters(self):
        """
        Return a string with the parameters serialized in a normalized order ready to be signed.
        """
        # serialize key-value pairs
        parameters = self.serialize_parameters()
        parameters.update(self._auth_header)  # add all oauth parameters but the signature
        # escape key-value pairs before sorting.
        key_values = [(escape(to_str(k)), escape(to_str(v))) for k, v in list(parameters.items())]
        # sort lexicographically
        key_values.sort()
        # combine key-value pairs
        return '&'.join(['%s=%s' % (k, v) for k, v in key_values])

    def normalize_url(self):
        """
        Parse the URL and rebuild it to be scheme://host/path.
        """
        parts = urllib.parse.urlparse(self._url)
        scheme, netloc, path = parts[:3]
        # exclude default port numbers.
        if scheme == 'http' and netloc[-3:] == ':80':
            netloc = netloc[:-3]
        elif scheme == 'https' and netloc[-4:] == ':443':
            netloc = netloc[:-4]
        return '%s://%s%s' % (scheme, netloc.lower(), path)

    def authorization_header(self, auth_header):
        """
        Return a list (signature, key-value mapping of authorization parameters).
        :param auth_header: the Authorization HTTP header
        """
        # check that the authorization header is OAuth
        if auth_header[:6] == 'OAuth ':
            auth_header = auth_header[6:]
            auth_header = self.parse_authorization_header(auth_header)
            for key in ('oauth_signature', 'oauth_timestamp'):
                if key not in auth_header:
                    raise AuthenticationError('Key "%s" is missing from Authorization header.' % key)
            header_sign = auth_header['oauth_signature']
            del auth_header['oauth_signature']  # not returned as part of authorization parameters
            return header_sign, auth_header

    @staticmethod
    def parse_authorization_header(header):
        """
        Parse Authorization header.
        """
        params = {}
        parts = header.split(',')
        for param in parts:
            # ignore realm
            if param.find('realm') > -1:
                continue
            # remove whitespace.
            param = param.strip()
            # split key-value
            param_parts = param.split('=', 1)
            # remove quotes and unescape value
            try:
                params[param_parts[0]] = urllib.parse.unquote(param_parts[1].strip('\"'))
            except IndexError:
                raise AuthenticationError('Invalid Authorization header.')
        return params


def escape(s):
    return urllib.parse.quote(s, safe='~')


def to_str(val):
    """
    Normalize parameter values to UTF-8 strings.
    """
    if isinstance(val, str):
        return val.encode("utf-8")
    else:
        return str(val)


class AuthenticationError(RuntimeError):
    def __init__(self, message):
        RuntimeError.__init__(self, message)
